# reports.pm - The Reporting Class to provide base functionality for Reports.
# Created by James Pattie, 11/11/1999.
# Last revised by Jason A. Pattie, 11/23/1999.
# Last Last revised by James Pattie, 01/10/2000.
# Updated to work with the new SubAccount Balances by James Pattie, 08/15/2000.
#$Id: reports.pm,v 1.3 2004/08/31 20:55:29 moreejt Exp $
#This package is released under the GNU General Public License.
#You should have recieved a copy of the GPL with it.
# Copyright (c) 2000 HLR http://www.pcxperience.org  All rights reserved.

use strict;
package reports;
use Pg;
use custom;

# Error variable for returning error messages from this class.
#$reports::ErrorString = "";  # initially empty
#$reports::ErrorOccurred = 0; # no error initially

# new - Expects the Class Object, Global Database pointer, Ledger Database name.
# You must check that no Error Occurred before you can assume that the object is complete.
sub new {
  my $class = shift;
  my $self = {};
  $self->{ErrorString} = "";
  $self->{ErrorOccurred} = 0;

  if (defined $_[0]) {
    $self->{GlobalDB} = shift;
  }
  else
  {
    $self->{ErrorString} .= "Error:  You must specify the Global Database Object to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0]) {
    $self->{LedgerDB} = shift;
  }
  else
  {
    $self->{ErrorString} .= "Error:  You must specify the Ledger Database Name to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  return bless $self, $class;
}

# getGlobalDB - Returns the pointer to the Global Database Object to work with.
sub getGlobalDB {
  my $self = shift;

  return $self->{GlobalDB};
}

# getLedgerDB - Returns the name of the Ledger Database we are working with.
sub getLedgerDB {
  my $self = shift;

  return $self->{LedgerDB};
}

# didErrorOccur - Returns the flag value to indicate if an error occurred.
sub didErrorOccur
{
  my $self = shift;

  return $self->{ErrorOccurred};
}

# getErrorMessage - Returns the Error Message created by the class.
sub getErrorMessage
{
  my $self = shift;

  return $self->{ErrorString};
}

# ClearErrorMessage - Will set the Error String back to nothing.
sub ClearErrorMessage
{
  my $self = shift;

  $self->{ErrorString} = "";
}

# getLedgerCode
#           Takes  : Nothing
#           Returns: The Code of the Ledger we are working with.
sub getLedgerCode
{
  my $self = shift;
  my $Result;
  my $result;
  my $conn = $self->getGlobalDB;
  my $ledger_db = $self->getLedgerDB;
  my $ledger_name;

  if (defined $_[0])
  {
    $self->{ErrorString} = "(getLedgerCode) Error:  This function does not take any arguments!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();  # Return an empty array to signal an error occurred.
  }

  # Get the Ledger Name
  $result = $conn->exec("SELECT name FROM entity_tb WHERE database_name = '$ledger_db'");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(getLedgerCode) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Ledger name for Ledger = '$ledger_db'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  $ledger_name = $result->getvalue(0, 0);
  #print "Ledger Name = '$ledger_name'\n";

  # Get the Ledger Code
  $result = $conn->exec("SELECT code FROM ledger_tb WHERE name = '$ledger_name'");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(getLedgerCode) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Ledger code for Ledger = '$ledger_db'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  $Result = $result->getvalue(0, 0);
  #print "Ledger Code = '$Result'\n";

  return $Result;
}

# getLedgerColor
#           Takes  : Nothing
#           Returns: The Color of the Ledger we are working with.
sub getLedgerColor
{
  my $self = shift;
  my $Result;
  my $result;
  my $conn = $self->getGlobalDB;
  my $ledger_db = $self->getLedgerDB;
  my $ledger_code;

  if (defined $_[0])
  {
    $self->{ErrorString} = "(getLedgerColor) Error:  This function does not take any arguments!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();  # Return an empty array to signal an error occurred.
  }

  # Get the Ledger Code
  $ledger_code = $self->getLedgerCode;
  if ($self->didErrorOccur())
  {
    return ();
  }

  $result = $conn->exec("SELECT color FROM ledger_tb WHERE code = '$ledger_code'");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(getLedgerColor) Error:  Getting the Ledger color failed!</h1>\n" . $conn->errorMessage . "<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  $Result = $result->getvalue(0, 0);

  return $Result;
}

# AccountActivity
#           Takes  : Array of Accounts with the Debit Increases Flag and Account Name
#           Takes  : Begin and End Dates
#	    Takes  : An optional flag that indicates the desired date search. 0=Exclude Both, 1=Include Both, 2=Exclude Begin and Include End.  Default = 1.
#           Returns: Array of activity and # of entries for each account
sub AccountActivity
{
  my $self = shift;
  my @Accounts;
  my @Results;
  my $begin_date;
  my $end_date;
  my $date_inclusive_search = 1;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(Account Activity) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Activity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Activity) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $date_inclusive_search = shift;
    if ($date_inclusive_search < 0 || $date_inclusive_search > 2)
    {
      $self->{ErrorString} .= "(Account Activity Split) Error:  Invalid value for the Inclusive Search Flag!\nMust be 0=Exclusive, 1=Inclusive or 2=Start Excluded, End Included.\n";
      $self->{ErrorOccurred} = 1;
    }
  }

  if ($self->{ErrorOccurred})
  {
    return ();  # Return an empty array to signal an error occurred.
  }

  #print "You specified " . (scalar @Accounts) . " Account(s) to get balances for.\n";

  # loop through the array of Accounts and for each Account lookup all transactions
  # that are within the date range and calculate the activity that took place using
  # the Debit Increases flag.  Should only have to work in the Ledger.
  my ($i, $acct_num, $debit_increases, $balance, $result, $acct_name);
  my $date_search_string = "";
  if ($date_inclusive_search == 1)
  {
    $date_search_string = "trans_date BETWEEN '$begin_date' AND '$end_date'";
  }
  elsif ($date_inclusive_search == 2)
  {
    $date_search_string = "trans_date > '$begin_date' AND trans_date <= '$end_date'";
  }
  elsif ($date_inclusive_search == 0)
  {
    $date_search_string = "trans_date > '$begin_date' AND trans_date < '$end_date'";
  }
  for ($i=0; $i < scalar @Accounts; $i++)
  {
    $acct_num = $Accounts[$i][0];
    $debit_increases = $Accounts[$i][1];
    $acct_name = $Accounts[$i][2];
    $balance = "0.00";

    $result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_num AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND $date_search_string)");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(Account Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transactions for Account ($acct_name) = '$acct_num', Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    my ($j, $amount, $dorc);
    for ($j=0; $j < $result->ntuples; $j++)
    {
      $amount = $result->getvalue($j, 0);
      # if we go back to using the money type, we will need to remove $ and ,'s from the amount.
      $dorc = $result->getvalue($j, 1);
      if ($debit_increases eq "t" && $dorc eq "t") # debit increases and amount is a debit
      {
        $balance += $amount;
      }
      elsif ($debit_increases eq "t" && $dorc eq "f") # debit increases and amount is a credit
      {
        $balance -= $amount;
      }
      elsif ($debit_increases eq "f" && $dorc eq "t") # debit decreases and amount is a debit
      {
        $balance -= $amount;
      }
      elsif ($debit_increases eq "f" && $dorc eq "f") # debit decreases and amount is a credit
      {
        $balance += $amount;
      }
    }

    $balance = sprintf "%.2f", $balance;

    my @Entry = ();
    $Entry[0] = $balance;
    $Entry[1] = $result->ntuples;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# AccountActivitySplit
#           Takes  : Array of Accounts with the Debit Increases Flag and Account Name
#           Takes  : Begin and End Dates
#	    Takes  : Optional Flag that indicates if dates are inclusive or not. Defaults to not being inclusive search. (0=non-inclusive, 1=inclusive, 2=exclude start include end)
#           Returns: Array of debits and credits
sub AccountActivitySplit
{
  my $self = shift;
  my @Accounts;
  my @Results;
  my $begin_date;
  my $end_date;
  my $date_inclusive_search = 0;  # Initially false.
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(Account Activity Split) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Activity Split) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Activity Split) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $date_inclusive_search = shift;
    if ($date_inclusive_search < 0 || $date_inclusive_search > 2)
    {
      $self->{ErrorString} .= "(Account Activity Split) Error:  Invalid value for the Inclusive Search Flag!\nMust be 0=Exclusive, 1=Inclusive or 2=Start Excluded, End Included.\n";
      $self->{ErrorOccurred} = 1;
    }
  }

  if ($self->{ErrorOccurred})
  {
    return ();  # Return an empty array to signal an error occurred.
  }

  #print "You specified " . (scalar @Accounts) . " Account(s) to get balances for.\n";

  # loop through the array of Accounts and for each Account lookup all transactions
  # that are within the date range and calculate the debits and credits activity that took place using
  # the Debit Increases flag.  Should only have to work in the Ledger.
  my ($i, $acct_num, $debit_increases, $debit_balance, $credit_balance, $result, $acct_name);
  my $date_search_string = "";
  if ($date_inclusive_search == 1)
  {
    $date_search_string = "trans_date BETWEEN '$begin_date' AND '$end_date'";
  }
  elsif ($date_inclusive_search == 2)
  {
    $date_search_string = "trans_date > '$begin_date' AND trans_date <= '$end_date'";
  }
  elsif ($date_inclusive_search == 0)
  {
    $date_search_string = "trans_date > '$begin_date' AND trans_date < '$end_date'";
  }
  for ($i=0; $i < scalar @Accounts; $i++)
  {
    $acct_num = $Accounts[$i][0];
    $debit_increases = $Accounts[$i][1];
    $acct_name = $Accounts[$i][2];
    $debit_balance = "0.00";
    $credit_balance = "0.00";

    # get all debits
    $result = $conn->exec("SELECT sum(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_num AND dorc = true AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND $date_search_string)");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(Account Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Sum of Debits for Account ($acct_name) = '$acct_num', Begin Date = '$begin_date', End Date = '$end_date', " . ($date_inclusive_search ? "Date Inclusive Search" : "Not Date Inclusive Search") . ".<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $debit_balance = $result->getvalue(0, 0);
    $debit_balance = sprintf "%.2f", $debit_balance;

    # get all credits
    $result = $conn->exec("SELECT sum(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_num AND dorc = false AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND $date_search_string)");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(Account Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Sum of Credits for Account ($acct_name) = '$acct_num', Begin Date = '$begin_date', End Date = '$end_date', " . ($date_inclusive_search ? "Date Inclusive Search" : "Not Date Inclusive Search") . ".<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $credit_balance = $result->getvalue(0, 0);
    $credit_balance = sprintf "%.2f", $credit_balance;

    my @Entry = ();
    $Entry[0] = $debit_balance;
    $Entry[1] = $credit_balance;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# GetAccountTypes
#    Takes  : first_digit (optional)
#    Returns: Array of Account Types.  All account types if first_digit not set, else just Account types of first_digit.
sub GetAccountTypes
{
  my $self = shift;
  my @Results = ();
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;
  my $first_digit = "";

  if (defined $_[0])
  {
    $first_digit = $_[0];
    shift;
    if ($first_digit < 0 || $first_digit > 5)
    {
      $self->{ErrorString} = "<h1>(GetAccountTypes) Error:  first_digit = '$first_digit' is invalid!<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
    }
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  # now we can do the actual lookup.
  if (length $first_digit > 0)
  {
    $result = $conn->exec("SELECT num FROM acct_type_tb WHERE first_digit = '$first_digit' ORDER BY num");
  }
  else
  {
    $result = $conn->exec("SELECT num FROM acct_type_tb ORDER BY num");
  }
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetAccountTypes) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Account Types in ledger = '$ledger_code'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  if ($result->ntuples > 0)
  {
    for (my $i=0; $i < $result->ntuples; $i++)
    {
      $Results[$i] = $result->getvalue($i, 0);
    }
  }
  else
  {
    $self->{ErrorString} = "<h1>(GetAccountTypes) Error:  No Account Types Returned!</h1>\n<br>Tried to get Account Types in ledger = '$ledger_code'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }

  return @Results;
}

# GetAccountTypeNames
#    Takes  : Array of Account Types
#    Returns: Array of Account Type Names in relation to the Account Types given.
sub GetAccountTypeNames
{
  my $self = shift;
  my @AccountTypes;
  my @Results;
  my $conn  = $self->getGlobalDB;

  if (defined $_[0])
  {
    @AccountTypes = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(GetAccountTypeNames) Error:  You must specify the Array of Account Types to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @AccountTypes) . " Account Types to work with.\n";

  my ($i, $acct_type_name, $result);
  for ($i=0; $i < scalar @AccountTypes; $i++)
  {
    $result = $conn->exec("SELECT name FROM acct_type_tb WHERE num = $AccountTypes[$i]");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(GetAccountTypeNames) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Account Type name for Account Type = '$AccountTypes[$i]'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $Results[$i] = $result->getvalue(0, 0);
  }

  return @Results;
}

# GetLocalAccounts
#    Takes  : Array of Account Types
#    Takes  : Taxable Flag (Optional, if not specified assumed to be 0) (0=don't care, 1=taxable only, 2=non-taxable only)
#    Returns: Array of Local Accounts for each Account Type with the Debit Increases Flag for each Account Type and the Account Name
#             Array [AccountType]
#             [AccountType] = [LocalAccounts]
#             [LocalAccounts] = Account Number, Debit Increases Flag, Account Name
sub GetLocalAccounts
{
  my $self = shift;
  my @AccountTypes;
  my @Results;
  my $Taxable;
  my $conn  = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    @AccountTypes = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(Get Local Accounts) Error:  You must specify the Array of Account Types to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $Taxable = shift;
  }
  else
  {  # if they don't specify the Taxable flag, then we default to not doing a tax special search.
    $Taxable = 0;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @AccountTypes) . " Account Types to work with.\n";
  #print "You want a " . $Taxable . " search conducted.\n";

  # loop through the array of AccountTypes and for each Account Type lookup all accounts
  # and return those that exist in the local_accounts_tb.  Also, return the Debit increases
  # value for each account based upon the Account Type info.
  my ($i, $acct_type_num, $debit_increases, $result);
  for ($i=0; $i < scalar @AccountTypes; $i++)
  {
    my @Info;  # stores the Accounts and Debit Increases values in and then gets put in the @Result array.
    $acct_type_num = $AccountTypes[$i];
    #print "Account Type=$AccountTypes[$i]\n";

    $result = $conn->exec("SELECT debit_increases FROM acct_type_tb WHERE num = $acct_type_num");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Getting Debit Increases value failed!</h1>\n" . $conn->errorMessage . "<br>Account Type = '$acct_type_num'<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();  # returns an empty array which is our signal to see if an error occurred.
    }
    if ($result->ntuples == 0)
    {
      $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Account Type is not configured!</h1>\nAccount Type = '$acct_type_num' does not appear to either exist or have it's debit increases field set.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $debit_increases = $result->getvalue(0, 0);

    # now we get the list of all accounts of this Account Type in the Global database
    # and then see how many of them are in the local_accounts_tb table in the Ledger.
    #$result = $conn->exec("SELECT number FROM account_tb WHERE type = $acct_type_num ORDER BY number");
    #if ($result->resultStatus != PGRES_TUPLES_OK)
    #{
    #  $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get list of Accounts for Account Type = '$acct_type_num'.<br>$sys_admin";
    #  $self->{ErrorOccurred} = 1;
    #  return ();
    #}

    #if ($result->ntuples > 0)
    #{
      my ($acct_num_list, $j, $acct_number, $acct_name);
      #for ($j=0; $j < $result->ntuples; $j++)
      #{
      #  $acct_number = $result->getvalue($j, 0);

      #  $acct_num_list .= ", " if ($j > 0);
      #  $acct_num_list .= $acct_number;
      #}

      # now get the accounts that exist in the local_accounts_tb of the ledger that match up.
      if ($Taxable == 1 && $acct_type_num != 15)
      {
        #$result = $conn->exec("SELECT acct_num FROM local_accounts_tb WHERE ledger_code = '$ledger_code' AND taxable = true AND sub_entity_code IS NULL AND acct_num IN ($acct_num_list) ORDER BY acct_num");
        $result = $conn->exec("SELECT la.acct_num FROM local_accounts_tb AS la, account_tb AS ac WHERE la.ledger_code = '$ledger_code' AND la.acct_num = ac.number AND ac.type = $acct_type_num AND la.taxable = true AND la.sub_entity_code IS NULL ORDER BY la.acct_num");
      }
      elsif ($Taxable == 2)
      {
        #print "Getting Non-Tax Accounts = '$acct_num_list'\n";
        #$result = $conn->exec("SELECT acct_num FROM local_accounts_tb WHERE ledger_code = '$ledger_code' AND taxable = false AND sub_entity_code IS NULL AND acct_num IN ($acct_num_list) ORDER BY acct_num");
        $result = $conn->exec("SELECT la.acct_num FROM local_accounts_tb AS la, account_tb AS ac WHERE la.ledger_code = '$ledger_code' AND la.acct_num = ac.number AND ac.type = $acct_type_num AND la.taxable = false AND la.sub_entity_code IS NULL ORDER BY la.acct_num");
      }
      else
      {
        #$result = $conn->exec("SELECT acct_num FROM local_accounts_tb WHERE ledger_code = '$ledger_code' AND sub_entity_code IS NULL AND acct_num IN ($acct_num_list) ORDER BY acct_num");
        $result = $conn->exec("SELECT la.acct_num FROM local_accounts_tb AS la, account_tb AS ac WHERE la.ledger_code = '$ledger_code' AND la.acct_num = ac.number AND ac.type = $acct_type_num AND la.sub_entity_code IS NULL ORDER BY acct_num");
      }
      if ($result->resultStatus != PGRES_TUPLES_OK)
      {
        $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get list of Accounts to work with based on the Local Accounts Table.<br>$sys_admin";
        $self->{ErrorOccurred} = 1;
        return ();
      }

      # now we walk the list and create the array that will make up this Account Type's entry.
      for ($j=0; $j < $result->ntuples; $j++)
      {
        my @Element;
        $acct_number = $result->getvalue($j, 0);

        # get this Accounts name
        my $result2 = $conn->exec("SELECT name FROM account_tb WHERE number = $acct_number");
        if ($result2->resultStatus != PGRES_TUPLES_OK)
        {
          $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Account Name for Account = '$acct_number'.<br>$sys_admin";
          $self->{ErrorOccurred} = 1;
          return ();
        }
        if ($result2->ntuples == 0)
        {
          $self->{ErrorString} = "<h1>(Get Local Accounts) Error:  Account no longer Exists!</h1>\nTried to get Account Name for Account = '$acct_number'.<br>$sys_admin";
          $self->{ErrorOccurred} = 1;
        }
        $acct_name = $result2->getvalue(0, 0);

        $Element[0] = $acct_number;
        $Element[1] = $debit_increases;
        $Element[2] = $acct_name;

        $Info[$j] = [ @Element ];
      }
    #}

    $Results[$i] = [ @Info ];
  }

  return @Results;
}

# AccountTypeActivity
#    Takes  : Array of Account Types
#    Takes  : begin and end dates
#    Takes  : Taxable Flag (Optional, if not specified assumed to be 0) (0=don't care, 1=only taxable, 2=only non-taxable)
#    Returns: Array of Activity and # of entries for each Account Type
sub AccountTypeActivity
{
  my $self = shift;
  my @AccountTypes;
  my @Results;
  my $Taxable;
  my $begin_date;
  my $end_date;

  if (defined $_[0])
  {
    @AccountTypes = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(Account Type Activity) Error:  You must specify the Array of Account Types to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Type Activity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Type Activity) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $Taxable = shift;
  }
  else
  {  # if they don't specify the Taxable flag, then we default to not doing a tax special search.
    $Taxable = 0;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @AccountTypes) . " Account Types to work with.\n";
  #print "You want a " . ($Taxable ? "Taxable":"Non-Taxable") . " search conducted.\n";
  #print "You specified Start Date = '$begin_date', End Date = '$end_date'\n";

  my ($i, $balance, $num_entries);
  my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  if ($self->{ErrorOccurred})
  {
    return ();
  }

  # now we need to walk through the AccountTypes array and for each entry,
  # sum up all the corresponding entries from the Accounts Array
  for ($i=0; $i < scalar @AccountTypes; $i++)
  {
    $balance = "0.00";
    $num_entries = 0;
    my @temp_array = ();
    my ($j);
    $temp_array[0] = $AccountTypes[$i];
    # Use GetLocalAccounts to get all Accounts for the specified Account Types
    my @Accounts = $this->GetLocalAccounts(\@temp_array, $Taxable);
    if ($self->{ErrorOccurred})
    {
      return ();  # just pass the error condition back to the user.
    }

    #print "GetLocalAccounts returned " . (scalar @{$Accounts[0]}) . " local Accounts for Account Type = '$AccountTypes[$i]'\n";
    #print "They are:\n";
    #for ($j=0; $j < scalar @{$Accounts[0]}; $j++)
    #{
    #  print "    Account = $Accounts[0][$j][0], Debit Increases = $Accounts[0][$j][1]\n";
    #}

    # now we get the Account activity using AccountActivity
    if (scalar @{$Accounts[0]} > 0)
    {
      my @accounts = ();
      for ($j=0; $j < scalar @{$Accounts[0]}; $j++)
      {
        my @temp;
        $temp[0] = $Accounts[0][$j][0];
        $temp[1] = $Accounts[0][$j][1];
        $temp[2] = $Accounts[0][$j][2];
        $accounts[$j] = [ @temp ];
      }

      my @balances = $this->AccountActivity(\@accounts, $begin_date, $end_date);
      if ($self->{ErrorOccurred})
      {
        return ();  # just pass the error condition back to the user.
      }

      # now we walk the balances array and make a sum of all balances encountered.
      for ($j=0; $j < scalar @balances; $j++)
      {
        $balance += $balances[$j][0];
        $num_entries += $balances[$j][1];
      }
    }

    $balance = sprintf "%.2f", $balance;  # make it look correct.

    my @Entry = ();
    $Entry[0] = $balance;
    $Entry[1] = $num_entries;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# GetSubAccounts
#    Takes  : Array of Accounts with Number, Debit Increases Value, Name
#    Returns: Array of SubAccounts for each Account (name and number)
#             Array = [SubAccount]
#             [SubAccount] = Number, Name
sub GetSubAccounts
{
  my $self = shift;
  my @Accounts;
  my @Results = ();
  my $conn = $self->getGlobalDB;

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(GetSubAccounts) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";

  my ($i, $j, $acct_number, $sub_acct_number, $result, $acct_name, $old_acct_type, $old_sub_account_types, @old_sub_accounts);

  $old_acct_type = -1;
  $old_sub_account_types = "";
  @old_sub_accounts = ();

  # now we need to walk through the Accounts array and for each entry,
  # get all the SubAccounts and their names
  for ($i=0; $i < scalar @Accounts; $i++)
  {
    $acct_number = $Accounts[$i][0]; # Debit Increases is at [1]
    $acct_name = $Accounts[$i][2];

    # find out the account type of this account.
    $result = $conn->exec("SELECT type FROM account_tb WHERE number = $acct_number");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(GetSubAccounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Account Type for Account ($acct_name) = '$acct_number'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    if ($result->ntuples == 0)
    {
      $self->{ErrorString} = "<h1>(GetSubAccounts) Error:  No Account Type Returned!</h1>\n<br>Tried to get Account Type for Account ($acct_name) = '$acct_number'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    my $acct_type = $result->getvalue(0, 0);
    my $sub_account_types = "";

    if ($acct_type != $old_acct_type)
    { # only lookup the account type sub account type info if not the same as the last account processed.
      $old_acct_type = $acct_type;
      # now get the Sub Account Types for this Account Type
      $result = $conn->exec("SELECT sub_acct_type FROM acct_type_tb WHERE num = $acct_type");
      if ($result->resultStatus != PGRES_TUPLES_OK)
      {
        $self->{ErrorString} = "<h1>(GetSubAccounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get SubAccountType for AccountType = '$acct_type'.<br>$sys_admin";
        $self->{ErrorOccurred} = 1;
        return ();
      }
      if ($result->ntuples == 0)
      {
        $self->{ErrorString} = "<h1>(GetSubAccounts) Error:  No SubAccountType Info Returned!</h1>\n<br>Tried to get SubAccountType for AccountType = '$acct_type'.<br>$sys_admin";
        $self->{ErrorOccurred} = 1;
        return ();
      }
      $sub_account_types = $result->getvalue(0, 0);
    }
    else
    {
      $sub_account_types = $old_sub_account_types;
    }

    my @temp_array = ();  # stores the SubAccounts info.

    if ($sub_account_types ne $old_sub_account_types)
    { # only lookup the sub accounts for these sub account types if not the same as the last run.
      # now we get all SubAccounts for this Account
      $result = $conn->exec("SELECT sub_acc_name, sub_acc_number FROM sub_acc_tb WHERE type IN ($sub_account_types) ORDER BY sub_acc_number, sub_acc_name");
      if ($result->resultStatus != PGRES_TUPLES_OK)
      {
        $self->{ErrorString} = "<h1>(GetSubAccounts) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get list of SubAccounts for Account ($acct_name) = '$acct_number'.<br>$sys_admin";
        $self->{ErrorOccurred} = 1;
        return ();
      }

      for ($j=0; $j < $result->ntuples; $j++)
      {
        my @Info = ();
        $Info[1] = $result->getvalue($j, 0);
        $Info[0] = $result->getvalue($j, 1);

        $temp_array[$j] = [ @Info ];
      }
      @old_sub_accounts = @temp_array;  # store the sub_accounts array.
    }
    else
    {
      @temp_array = @old_sub_accounts;
    }

    $Results[$i] = [ @temp_array ];
  }

  return @Results;
}

# SubAccountActivity
#    Takes  : Array of Accounts with Debit Increases value
#             Array [Accounts]
#             [Accounts] = [Account Info]
#             [Account Info] = Account Number, Debit Increases Value, Account Name
#    Takes  : For each Account, an Array of SubAccounts
#             Array [Accounts]
#             [Accounts] = [SubAccounts]
#             [SubAccounts] = SubAccount Number, SubAccount Name
#    Takes  : begin and end dates
#    Returns: Array of Balances, with an array of SubAccount balances for each Account.
#             Array [Accounts]
#             [Accounts] = balance_total, [SubAccounts Balance], total # entries
#             [SubAccounts Balance] = balance of SubAccount and # of entries in same order as in SubAccount array.
sub SubAccountActivity
{
  my $self = shift;
  my @Accounts;
  my @SubAccounts;
  my @Results;
  my $begin_date;
  my $end_date;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;
  my $result;

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(SubAccount Activity) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @SubAccounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubAccount Activity) Error:  You must specify the Array of SubAccounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubAccount Activity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubAccount Activity) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";
  #print "You specified Start Date = '$begin_date', End Date = '$end_date'\n";

  my ($i, $j, $balance, $sub_balance, $acct_number, $debit_increases, $dorc, $acct_name, $amount, $total_entries);
  #my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  #if ($self->{ErrorOccurred})
  #{
  #  return ();
  #}

  # now we need to walk through the Accounts array and for each entry,
  # get all the SubAccount balances and sum them up.
  for ($i=0; $i < scalar @Accounts; $i++)
  {
    $balance = "0.00";
    $acct_number = $Accounts[$i][0];
    $debit_increases = $Accounts[$i][1];
    $acct_name = $Accounts[$i][2];
    $total_entries = 0;
    #print "Account ($acct_name) = $acct_number\n";

    # now we get the SubAccount activity.
    my @sub_array = ();
    if (scalar @{$SubAccounts[$i]} > 0)
    {
      #print "You gave me " . (scalar @{$SubAccounts[$i]}) . " SubAccounts to work with for index $i.\n";
      # for each Sub Account, get the balance of all entries it is in.
      for ($j=0; $j < scalar @{$SubAccounts[$i]}; $j++)
      {
        $sub_balance = "0.00";
        my $sub_acct_number = $SubAccounts[$i][$j][0];
        my $sub_acct_name   = $SubAccounts[$i][$j][1];
        #print "Sub-Account ($sub_acct_name) = $sub_acct_number\n";
        $result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND sub_account_num = $sub_acct_number AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND (trans_date BETWEEN '$begin_date' AND '$end_date'))");
        if ($result->resultStatus != PGRES_TUPLES_OK)
        {
          $self->{ErrorString} = "<h1>(Account Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transactions for Account ($acct_name) = '$acct_number', SubAccount ($sub_acct_name) = '$sub_acct_number', Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
          $self->{ErrorOccurred} = 1;
          return ();
        }
        my ($k);
        for ($k=0; $k < $result->ntuples; $k++)
        {
          $amount = $result->getvalue($k, 0);
          $dorc = $result->getvalue($k, 1);
          #print "Amount = $amount, DORC = $dorc\n";
          #print "Debit Increases = $debit_increases\n";
          if ($debit_increases eq "t" && $dorc eq "t") # debit increases and amount is a debit
          {
            $sub_balance += $amount;
          }
          elsif ($debit_increases eq "t" && $dorc eq "f") # debit increases and amount is a credit
          {
            $sub_balance -= $amount;
          }
          elsif ($debit_increases eq "f" && $dorc eq "t") # debit decreases and amount is a debit
          {
            $sub_balance -= $amount;
          }
          elsif ($debit_increases eq "f" && $dorc eq "f") # debit decreases and amount is a credit
          {
            $sub_balance += $amount;
          }
          #print "Sub_Balance = $sub_balance\n";
        }

        $sub_balance = sprintf "%.2f", $sub_balance;

        $balance += $sub_balance;

        my @Entry = ();
        $Entry[0] = $sub_balance;
        $Entry[1] = $result->ntuples;
        $total_entries += $result->ntuples;

        $sub_array[$j] = [ @Entry ];
      }
    }
    $balance = sprintf "%.2f", $balance;  # make it look correct.

    my @Entry = ();

    $Entry[0] = $balance;
    $Entry[1] = [ @sub_array ];
    $Entry[2] = $total_entries;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# GetSubEntities
#    Takes  : Nothing
#    Returns: Array of SubEntities in the current Ledger
#             Array = [SubEntities]
#             [SubEntities] = Code, Name
sub GetSubEntities
{
  my $self = shift;
  my @Results = ();
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    $self->{ErrorString} = "(GetSubEntities) Error:  This function does not take any input!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $result);

  # get all the SubEntities and their codes
  $result = $conn->exec("SELECT code, name FROM sub_entity_tb WHERE ledger_code = '$ledger_code' ORDER BY name, code");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetSubEntities) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get All SubEntities from the Ledger.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  # Insert at the head the Null Sub Entity that is used by default (Code = '', Name = '');
  my @temp_array = ();
  $temp_array[0] = '';
  $temp_array[1] = '';
  $Results[0] = [ @temp_array ];

  for ($i=0; $i < $result->ntuples; $i++)
  {
    my @temp_array = ();
    my $sub_e_code = $result->getvalue($i, 0);
    my $sub_e_name = $result->getvalue($i, 1);

    $temp_array[0] = $sub_e_code;
    $temp_array[1] = $sub_e_name;

    $Results[$i+1] = [ @temp_array ];
  }

  return @Results;
}

# SubEntityAccountActivity
#    Takes  : Array of SubEntities (Codes, Names)
#    Takes  : Array of Accounts with Debit Increases value
#             Array [Accounts]
#             [Accounts] = [Account Info]
#             [Account Info] = Account Number, Debit Increases Value, Account Name
#    Takes  : begin and end dates
#    Returns: Array of Balances, with an array of Account balances for each SubEntity.
#             Array [SubEntities]
#             [SubEntities] = balance_total, [Accounts Balance], num_entries
#             [Accounts Balance] = Activity of Account and # of entries in same order as in Account array.
sub SubEntityAccountActivity
{
  my $self = shift;
  my @Accounts = ();
  my @SubEntities = ();
  my @Results = ();
  my $begin_date;
  my $end_date;
  my $conn = $self->getGlobalDB;
  my $result;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    @SubEntities = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(SubEntityAccount Activity) Error:  You must specify the Array of SubEntities to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntityAccount Activity) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntityAccount Activity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntityAccount Activity) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";
  #print "You specified Start Date = '$begin_date', End Date = '$end_date'\n";

  my ($i, $j, $balance, $sub_balance, $acct_number, $debit_increases, $dorc, $acct_name, $amount, $sub_entity_code, $sub_entity_name, $num_entries);
  #my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  #if ($self->{ErrorOccurred})
  #{
  #  return ();
  #}

  # now we need to walk through the SubEntities array and for each entry,
  # get all the Account balances and sum them up.
  for ($i=0; $i < scalar @SubEntities; $i++)
  {
    $balance = "0.00";
    $num_entries = 0;
    $sub_entity_name = $SubEntities[$i][1];
    $sub_entity_code = $SubEntities[$i][0];
    my @sub_array = ();

    #print "SubEntity Name = $sub_entity_name, Code = $sub_entity_code\n";
    #print "There are " . (scalar @Accounts) . " Accounts to work with.\n";

    for ($j=0; $j < scalar @Accounts; $j++)
    {
      $acct_number = $Accounts[$j][0];
      $debit_increases = $Accounts[$j][1];
      $acct_name = $Accounts[$j][2];

      #print "Account ($acct_name) = $acct_number, Debit Increases = $debit_increases\n";

      # now we get the Account activity for this Account and SubEntity combo.
      $sub_balance = "0.00";
      $result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND sub_e_code = '$sub_entity_code' AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND trans_date BETWEEN '$begin_date' AND '$end_date')");
      if ($result->resultStatus != PGRES_TUPLES_OK)
      {
        $self->{ErrorString} = "<h1>(SubEntityAccount Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transactions for Account ($acct_name) = '$acct_number', SubEntity ($sub_entity_name) = '$sub_entity_code', Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
        $self->{ErrorOccurred} = 1;
        return ();
      }
      my ($k);
      #print $result->ntuples . " entries returned.\n";
      for ($k=0; $k < $result->ntuples; $k++)
      {
        $amount = $result->getvalue($k, 0);
        $dorc = $result->getvalue($k, 1);
        #print "Amount = $amount, DORC = $dorc\n";
        #print "Debit Increases = $debit_increases\n";
        if ($debit_increases eq "t" && $dorc eq "t") # debit increases and amount is a debit
        {
          $sub_balance += $amount;
        }
        elsif ($debit_increases eq "t" && $dorc eq "f") # debit increases and amount is a credit
        {
          $sub_balance -= $amount;
        }
        elsif ($debit_increases eq "f" && $dorc eq "t") # debit decreases and amount is a debit
        {
          $sub_balance -= $amount;
        }
        elsif ($debit_increases eq "f" && $dorc eq "f") # debit decreases and amount is a credit
        {
          $sub_balance += $amount;
        }
        #print "Sub_Balance = $sub_balance\n";
      }

      $sub_balance = sprintf "%.2f", $sub_balance;

      $balance += $sub_balance;

      my @Entry = ();
      $Entry[0] = $sub_balance;
      $Entry[1] = $result->ntuples;
      $num_entries += $result->ntuples;

      $sub_array[$j] = [ @Entry ];
    }
    $balance = sprintf "%.2f", $balance;  # make it look correct.

    my @Entry = ();

    $Entry[0] = $balance;
    $Entry[1] = [ @sub_array ];
    $Entry[2] = $num_entries;  # store the number of entries that made up this balance.

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# SubEntitySubAccountActivity
#    Takes  : Array of SubEntities (Codes, Names)
#    Takes  : Array of Accounts with Debit Increases value
#             Array [Accounts]
#             [Accounts] = [Account Info]
#             [Account Info] = Account Number, Debit Increases Value, Account Name
#    Takes  : For each Account, an Array of SubAccounts
#             Array [Accounts]
#             [Accounts] = [SubAccounts]
#             [SubAccounts] = SubAccount Number, SubAccount Name
#    Takes  : begin and end dates
#    Returns: Array of Balances for each SubEntity, with an array of SubAccount balances for each Account.
#             Array [SubEntities] = balance_total, [Accounts], # of entries for this SubEntity
#             [Accounts] = balance_total, [SubAccounts Balance], # of entries for this Account
#             [SubAccounts Balance] = balance of SubAccount and # of entries in same order as in SubAccount array.
sub SubEntitySubAccountActivity
{
  my $self = shift;
  my @Accounts;
  my @SubAccounts;
  my @SubEntities = ();
  my @Results;
  my $begin_date;
  my $end_date;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;
  my $result;

  if (defined $_[0])
  {
    @SubEntities = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(SubEntitySubAccountActivity) Error:  You must specify the Array of SubEntities to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @Accounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntitySubAccountActivity) Error:  You must specify the Array of Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @SubAccounts = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntitySubAccountActivity) Error:  You must specify the Array of SubAccounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntitySubAccountActivity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(SubEntitySubAccountActivity) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";
  #print "You specified Start Date = '$begin_date', End Date = '$end_date'\n";

  my ($i, $j, $l, $balance, $sub_balance, $acct_number, $debit_increases, $dorc, $acct_name, $amount);
  #my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  #if ($self->{ErrorOccurred})
  #{
  #  return ();
  #}

  # now we need to walk through the SubEntities array and for each entry,
  # go through the Accounts array and do stuff.
  for ($l=0; $l < scalar @SubEntities; $l++)
  {
    my $sub_entity_code = $SubEntities[$l][0];
    my $sub_entity_name = $SubEntities[$l][1];
    my $total_balance = "0.00";
    my $total_entries = 0;
    my @account_array = ();

    # now we need to walk through the Accounts array and for each entry,
    # get all the SubAccount balances and sum them up.
    for ($i=0; $i < scalar @Accounts; $i++)
    {
      $balance = "0.00";
      $acct_number = $Accounts[$i][0];
      $debit_increases = $Accounts[$i][1];
      $acct_name = $Accounts[$i][2];
      my $num_entries = 0;

      # now we get the SubAccount activity.
      my @sub_array = ();
      #print "\$i = $i\n";
      if (scalar @{$SubAccounts[$i]} > 0)
      {
        #print "You gave me " . (scalar @{$SubAccounts[$i]}) . " SubAccounts to work with for index $i.\n";
        # for each Sub Account, get the balance of all entries it is in.
        for ($j=0; $j < scalar @{$SubAccounts[$i]}; $j++)
        {
          $sub_balance = "0.00";
          my $sub_acct_number = $SubAccounts[$i][$j][0];
          my $sub_acct_name   = $SubAccounts[$i][$j][1];
          #print "Sub-Account ($sub_acct_name) = $sub_acct_number\n";
          $result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND sub_account_num = $sub_acct_number AND sub_e_code = '$sub_entity_code' AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND trans_date BETWEEN '$begin_date' AND '$end_date')");
          if ($result->resultStatus != PGRES_TUPLES_OK)
          {
            $self->{ErrorString} = "<h1>(Account Activity) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transactions for Account ($acct_name) = '$acct_number', SubAccount ($sub_acct_name) = '$sub_acct_number', SubEntity ($sub_entity_name) = $sub_entity_code, Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
            $self->{ErrorOccurred} = 1;
            return ();
          }
          my ($k);
          for ($k=0; $k < $result->ntuples; $k++)
          {
            $amount = $result->getvalue($k, 0);
            $dorc = $result->getvalue($k, 1);
            #print "Amount = $amount, DORC = $dorc\n";
            #print "Debit Increases = $debit_increases\n";
            if ($debit_increases eq "t" && $dorc eq "t") # debit increases and amount is a debit
            {
              $sub_balance += $amount;
            }
            elsif ($debit_increases eq "t" && $dorc eq "f") # debit increases and amount is a credit
            {
              $sub_balance -= $amount;
            }
            elsif ($debit_increases eq "f" && $dorc eq "t") # debit decreases and amount is a debit
            {
              $sub_balance -= $amount;
            }
            elsif ($debit_increases eq "f" && $dorc eq "f") # debit decreases and amount is a credit
            {
              $sub_balance += $amount;
            }
            #print "Sub_Balance = $sub_balance\n";
          }

          $sub_balance = sprintf "%.2f", $sub_balance;

          $balance += $sub_balance;
          $num_entries += $result->ntuples;

          my @Entry = ();
          $Entry[0] = $sub_balance;
          $Entry[1] = $result->ntuples;

          $sub_array[$j] = [ @Entry ];
        }
      } # end of sub-account for loop
      $balance = sprintf "%.2f", $balance;  # make it look correct.

      my @Entry = ();

      $Entry[0] = $balance;
      $Entry[1] = [ @sub_array ];
      $Entry[2] = $num_entries;

      $total_balance += $balance;
      $total_entries += $num_entries;

      $account_array[$i] = [ @Entry ];
    } # end of Account for loop

    my @Entry = ();

    $total_balance = sprintf "%.2f", $total_balance;

    $Entry[0] = $total_balance;
    $Entry[1] = [ @account_array ];
    $Entry[2] = $total_entries;

    $Results[$l] = [ @Entry ];
  } # end of SubEntities for loop

  return @Results;
}

# GetActivityData
#    Takes  : Account Info Array (Number, Debit Increases Value, Name) [Required]
#    Takes  : SubAccount Array (SubAccount Number, SubAccount Name) [Optional may be ()]
#    Takes  : SubEntity Array (Code, Name) [Optional may be ()]
#    Takes  : begin and end dates [Required]
#    Returns: Array of Entries from the debitcredit_tb that match the given data.
#             Array [Entries]
#             [Entries] = parent_num, sub_e_code, account_num, sub_account_num, sub_account_name, amount, dorc, cleared, counter, trans_date, payee_payer, memo, journal
sub GetActivityData
{
  my $self = shift;
  my @Account = ();
  my @SubAccount = ();
  my @SubEntity = ();
  my @Results = ();
  my $begin_date;
  my $end_date;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    @Account = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(GetActivityData) Error:  You must specify the Account to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @SubAccount = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(GetActivityData) Error:  You must specify the SubAccount to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    @SubEntity = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} .= "(GetActivityData) Error:  You must specify the SubEntity to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(GetActivityData) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $end_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(GetActivityData) Error:  You must specify the Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";
  #print "You specified Start Date = '$begin_date', End Date = '$end_date'\n";

  my ($i, $j, $acct_number, $debit_increases, $dorc, $acct_name, $amount, $sub_acct_number, $sub_acct_name, $sub_entity_code, $sub_entity_name, $cleared, $counter, $parent_num, $result, $other_search_info, $result2, $trans_date, $memo, $journal, $payee_payer);
  #my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  #if ($self->{ErrorOccurred})
  #{
  #  return ();
  #}

  # now we need to get the Account info
  $acct_number = $Account[0];
  $acct_name = $Account[2];
  $debit_increases = $Account[1];

  # next see if there is any sub-account info to use.
  if (scalar @SubAccount == 2)
  {
    $sub_acct_number = $SubAccount[0];
    $sub_acct_name = $SubAccount[1];
    $other_search_info = "AND sub_account_num = $sub_acct_number";
  }

  # next see if there is any sub-entity info to use.
  if (scalar @SubEntity == 2)
  {
    $sub_entity_code = $SubEntity[0];
    $sub_entity_name = $SubEntity[1];
    $other_search_info .= " AND sub_e_code = '$sub_entity_code'";
  }

  # now we can do the actuall lookup.
  #print $other_search_info;

  # now we get the transaction information for this entry
  $result = $conn->exec("SELECT ref_num, trans_date, payee_payer, memo, journal_code FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND (trans_date BETWEEN '$begin_date' AND '$end_date') AND ref_num IN (SELECT parent_num FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number $other_search_info) ORDER BY trans_date, ref_num");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetActivityData) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get list of Transaction Info for Account ($acct_name) = '$acct_number', SubAccount ($sub_acct_name) = '$sub_acct_number', SubEntity ($sub_entity_name) = '$sub_entity_code', Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  my $count = 0;
  for (my $k=0; $k < $result->ntuples; $k++)
  {
    $parent_num = $result->getvalue($k, 0);
    $trans_date = $result->getvalue($k, 1);
    $payee_payer = $result->getvalue($k, 2);
    $memo = $result->getvalue($k, 3);
    $journal = $result->getvalue($k, 4);

    my $result3 = $conn->exec("SELECT sub_e_code, account_num, sub_account_num, amount, dorc, cleared, counter FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND parent_num = $parent_num");
    if ($result3->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(GetActivityData) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transaction info for Account ($acct_name) = '$acct_number', SubAccount ($sub_acct_name) = '$sub_acct_number', SubEntity ($sub_entity_name) = '$sub_entity_code', Begin Date = '$begin_date', End Date = '$end_date'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }

    #print $result2->ntuples . " entries returned.\n";
    for (my $l=0; $l < $result3->ntuples; $l++)
    {
      $sub_entity_code = $result3->getvalue($l, 0);
      $acct_number = $result3->getvalue($l, 1);
      $sub_acct_number = $result3->getvalue($l, 2);
      $amount = $result3->getvalue($l, 3);
      $dorc = $result3->getvalue($l, 4);
      $cleared = $result3->getvalue($l, 5);
      $counter = $result3->getvalue($l, 6);
      #print "Amount = $amount, DORC = $dorc\n";

      if ($sub_acct_number != -1) # lookup the name that goes with it.
      {
        # first verify that we are not working with a cash account before we do this.
        $result2 = $conn->exec("SELECT type FROM account_tb WHERE number = $acct_number");
        if ($result2->resultStatus != PGRES_TUPLES_OK)
        {
          $self->{ErrorString} = "<h1>(GetActivityData) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Account Type info for Account = '$acct_number'.<br>$sys_admin";
          $self->{ErrorOccurred} = 1;
          return ();
        }
        if ($result2->getvalue(0, 0) != 0)
        {
          $result2 = $conn->exec("SELECT sub_acc_name FROM sub_acc_tb WHERE sub_acc_number = $sub_acct_number AND parent = $acct_number");
          if ($result2->resultStatus != PGRES_TUPLES_OK)
          {
            $self->{ErrorString} = "<h1>(GetActivityData) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Sub Account Name for Sub Account = '$sub_acct_number'.<br>$sys_admin";
            $self->{ErrorOccurred} = 1;
            return ();
          }
          if ($result2->ntuples != 1)
          {
            $self->{ErrorString} = "<h1>(GetActivityData) Error:  No Sub Account Name Exists!</h1>\nTried to get Sub Account Name for Sub Account = '$sub_acct_number'.<br>$sys_admin";
            $self->{ErrorOccurred} = 1;
            return ();
          }

          $sub_acct_name = $result2->getvalue(0, 0);
        }
      }

      my @Entry = ();

      $Entry[0] = $parent_num;
      $Entry[1] = $sub_entity_code;
      $Entry[2] = $acct_number;
      $Entry[3] = $sub_acct_number;
      $Entry[4] = $sub_acct_name;
      $Entry[5] = $amount;
      $Entry[6] = $dorc;
      $Entry[7] = $cleared;
      $Entry[8] = $counter;
      $Entry[9] = $trans_date;
      $Entry[10] = $payee_payer;
      $Entry[11] = $memo;
      $Entry[12] = $journal;

      $Results[$count++] = [ @Entry ];
    }
  }

  return @Results;
}

# GetAccountBalances
#    Takes  : Account Info Array (Number, Debit Increases Value, Name) [Required]
#    Takes  : SubAccount (Number) [Optional]
#    Takes  : date to calculate balance to [Required]
#    Returns: Array of Balances and # of entries for all accounts specified.
sub GetAccountBalances
{
  my $self = shift;
  my @Accounts = ();
  my $begin_date = "";
  my @Results = ();
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;
  my $sub_account = 0;
  my $sub_acct_string;

  if (!defined $_[0])
  {
    $self->{ErrorString} = "(GetAccountBalances) Error:  You must specify the Accounts to work with!\n";
    $self->{ErrorOccurred} = 1;
  }
  else
  {
    @Accounts = @{$_[0]};
    shift;
  }

  if (!defined $_[0])
  {
    $self->{ErrorString} .= "(GetAccountBalances) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }
  else
  {
    $begin_date = $_[0];
    shift;
  }

  if (defined $_[0])
  {
    $sub_account = $_[0];
    shift;
    $sub_acct_string = "AND db.sub_account_num = $sub_account";
  }
  else
  {
    $sub_account = -2;
    $sub_acct_string = "";
  }

  if ($self->{ErrorOccurred})
  {
    $self->{ErrorString} .= "(GetAccountBalances) \$begin_date = '$begin_date', \@Accounts = '@Accounts', \$sub_account = '$sub_account'\n";

    return ();
  }

  #print "You gave me " . (scalar @Accounts) . " Accounts to work with.\n";
  #print "You specified Start Date = '$begin_date'\n";

  my ($i, $j, $acct_number, $debit_increases, $dorc, $acct_name, $amount, $result);

  for ($i=0; $i < scalar @Accounts; $i++)
  {
    my $temp_bal = "0.00";
    my $temp_date = "Never";
    # now we need to get the Account info
    $acct_number = $Accounts[$i][0];
    $acct_name = $Accounts[$i][2];
    $debit_increases = $Accounts[$i][1];

    # now we can do the actual lookup.
    $result = $conn->exec("SELECT balance, bal_date FROM acct_bal_tb WHERE ledger_code = '$ledger_code' AND acct_num = $acct_number AND sub_acct_num = $sub_account AND bal_date <= '$begin_date' ORDER BY bal_date DESC");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(GetAccountBalances) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get Balance Table Entry for Account ($acct_name) = '$acct_number', Sub Account = '$sub_account' on or before Date = '$begin_date'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    if ($result->ntuples > 0)
    {
      $temp_bal = $result->getvalue(0, 0);
      $temp_date = $result->getvalue(0, 1);

      if (not $temp_date eq $begin_date)
      {
        #$result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND (trans_date > '$temp_date' AND trans_date <= '$begin_date'))");
        $result = $conn->exec("SELECT db.amount, db.dorc FROM debitcredit_tb AS db WHERE db.ledger_code = '$ledger_code' AND db.account_num = $acct_number AND (db.trans_date > '$temp_date' AND db.trans_date <= '$begin_date') $sub_acct_string");
      }
    }
    else
    {
      #print "No previous balance encountered for account $acct_number!\n";
      #$result = $conn->exec("SELECT amount, dorc FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND account_num = $acct_number AND parent_num IN (SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND trans_date <= '$begin_date')");
      $result = $conn->exec("SELECT db.amount, db.dorc FROM debitcredit_tb AS db WHERE db.ledger_code = '$ledger_code' AND db.account_num = $acct_number AND (db.trans_date <= '$begin_date') $sub_acct_string");
    }
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(GetAccountBalances) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get transactions for Account ($acct_name) = $acct_number, Sub Account = '$sub_account', Start Date = '$temp_date', Stop Date = '$begin_date'.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    if ($result->ntuples > 0)
    {
      for ($j=0; $j < $result->ntuples; $j++)
      {
        $amount = $result->getvalue($j, 0);
        $dorc = $result->getvalue($j, 1);
        if ($debit_increases eq "t" && $dorc eq "t") # debit increases and amount is a debit
        {
          $temp_bal += $amount;
        }
        elsif ($debit_increases eq "t" && $dorc eq "f") # debit increases and amount is a credit
        {
          $temp_bal -= $amount;
        }
        elsif ($debit_increases eq "f" && $dorc eq "t") # debit decreases and amount is a debit
        {
          $temp_bal -= $amount;
        }
        elsif ($debit_increases eq "f" && $dorc eq "f") # debit decreases and amount is a credit
        {
          $temp_bal += $amount;
        }
        $temp_bal = sprintf "%.2f", $temp_bal;
      }
    }
#    else
#    {
#      print "No amounts returned for account $acct_number!\n";
#    }

    #$temp_bal = sprintf "%.2f", $temp_bal;

    my @Entry = ();
    $Entry[0] = $temp_bal;
    $Entry[1] = $result->ntuples;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# GetAccountTypeBalances
#    Takes  : Array of Account Types
#    Takes  : Date to get the balance up till
#    Takes  : Taxable Flag (Optional, if not specified assumed to be 0)
#    Returns: Array of Balances and # of entries for each Account Type
sub GetAccountTypeBalances
{
  my $self = shift;
  my @AccountTypes;
  my @Results;
  my $Taxable;
  my $begin_date;

  if (defined $_[0])
  {
    @AccountTypes = @{$_[0]};
    shift;
  }
  else
  {
    $self->{ErrorString} = "(Account Type Activity) Error:  You must specify the Array of Account Types to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $begin_date = shift;
  }
  else
  {
    $self->{ErrorString} .= "(Account Type Activity) Error:  You must specify the Beginning Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $Taxable = shift;
  }
  else
  {  # if they don't specify the Taxable flag, then we default to not doing a tax special search.
    $Taxable = 0;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $balance, $num_entries);
  my $this = reports->new($self->getGlobalDB, $self->getLedgerDB);
  if ($self->{ErrorOccurred})
  {
    return ();
  }

  # now we need to walk through the AccountTypes array and for each entry,
  # sum up all the corresponding entries from the Accounts Array
  for ($i=0; $i < scalar @AccountTypes; $i++)
  {
    $balance = "0.00";
    $num_entries = 0;
    my @temp_array = ();
    my ($j);
    $temp_array[0] = $AccountTypes[$i];
    # Use GetLocalAccounts to get all Accounts for the specified Account Types
    my @Accounts = $this->GetLocalAccounts(\@temp_array, $Taxable);
    if ($self->{ErrorOccurred})
    {
      return ();  # just pass the error condition back to the user.
    }

    # now we get the Account activity using AccountActivity
    if (scalar @{$Accounts[0]} > 0)
    {
      #print "(GetAccountTypeBalances) We have some Accounts to work with for AccountType = $AccountTypes[$i].\n";
      my @accounts = ();
      for ($j=0; $j < scalar @{$Accounts[0]}; $j++)
      {
        my @temp;
        $temp[0] = $Accounts[0][$j][0];
        $temp[1] = $Accounts[0][$j][1];
        $temp[2] = $Accounts[0][$j][2];
        $accounts[$j] = [ @temp ];
      }

      my @balances = $this->GetAccountBalances(\@accounts, $begin_date);
      if ($self->{ErrorOccurred})
      {
        return ();  # just pass the error condition back to the user.
      }

      # now we walk the balances array and make a sum of all balances encountered.
      for ($j=0; $j < scalar @balances; $j++)
      {
        #print "(GetAccountTypeBalances) Balance returned for Account=$Accounts[0][$j][2] = $balances[$j][0], # = $balances[$j][1].<br>\n";
        $balance += $balances[$j][0];
        $num_entries += $balances[$j][1];
      }
    }

    $balance = sprintf "%.2f", $balance;  # make it look correct.

    my @Entry = ();
    $Entry[0] = $balance;
    $Entry[1] = $num_entries;

    $Results[$i] = [ @Entry ];
  }

  return @Results;
}

# DoesSystemBalance
#    Takes  : Start and End Dates to search across
#    Returns: Array with debit balance and credit balance
sub DoesSystemBalance
{
  my $self = shift;
  my @Results = ();
  my $Start_Date;
  my $End_Date;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    $Start_Date = shift;
  }
  else
  {
    $self->{ErrorString} = "(DoesSystemBalance) Error:  You must specify a Starting Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $End_Date = shift;
  }
  else
  {
    $self->{ErrorString} = "(DoesSystemBalance) Error:  You must specify a Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $result, $result2, $ref_nums);

  # initialize the Results array
  $Results[0] = "0.00";
  $Results[1] = "0.00";

  # get the list of all transactions
  #$result2 = $conn->exec("SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND (trans_date BETWEEN '$Start_Date' AND '$End_Date') ORDER BY ref_num");
  #if ($result2->resultStatus != PGRES_TUPLES_OK)
  #{
  #  $self->{ErrorString} = "<h1>(DoesSystemBalance) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get all Reference Numbers of Transactions between $Start_Date and $End_Date from the Ledger.<br>$sys_admin";
  #  $self->{ErrorOccurred} = 1;
  #  return ();
  #}

  #for ($i=0; $i < $result2->ntuples; $i++)
  #{
  #  my $ref_num = $result2->getvalue($i, 0);
  #  $ref_nums .= "," if ($i > 0);
  #  $ref_nums .= $ref_num;
  #}

  #if ($result2->ntuples > 0)
  #{
    # get the sum of all debits and all credits for this ref_num and verify that they equal
    #$result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND dorc = true AND parent_num IN ($ref_nums)"); # all debits
    $result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb AS db WHERE db.ledger_code = '$ledger_code' AND (db.trans_date BETWEEN '$Start_Date' AND '$End_Date') AND dorc = true");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(DoesSystemBalance) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get sum of all Debits from the Ledger.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    my $debit = $result->getvalue(0, 0);
    $Results[0] += $debit;

    #$result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND dorc = false AND parent_num IN ($ref_nums)"); # all credits
    $result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb AS db WHERE db.ledger_code = '$ledger_code' AND (db.trans_date BETWEEN '$Start_Date' AND '$End_Date') AND dorc = false");
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(DoesSystemBalance) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get sum of all Credits from the Ledger.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    my $credit = $result->getvalue(0, 0);
    $Results[1] += $credit;
  #}

  $Results[0] = sprintf "%.2f", $Results[0];
  $Results[1] = sprintf "%.2f", $Results[1];

  if ($Results[0] eq $Results[1])
  {
    $Results[2] = 1;
  }
  else
  {
    $Results[2] = 0;
  }

  return @Results;
}

# FindInvalidEntries
#    Takes  : Start and End Dates to search across
#    Returns: Array of ref_num's of all invalid Transactions in the debitcredit_tb.
sub FindInvalidEntries
{
  my $self = shift;
  my @Results = ();
  my $Start_Date;
  my $End_Date;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    $Start_Date = shift;
  }
  else
  {
    $self->{ErrorString} = "(FindInvalidEntries) Error:  You must specify a Starting Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if (defined $_[0])
  {
    $End_Date = shift;
  }
  else
  {
    $self->{ErrorString} = "(FindInvalidEntries) Error:  You must specify a Ending Date to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $result, $result2, $debit, $credit, $ref_num, $j);

  # get the list of all transactions
  $result2 = $conn->exec("SELECT ref_num FROM journal_entry_tb WHERE ledger_code = '$ledger_code' AND (trans_date BETWEEN '$Start_Date' AND '$End_Date')");
  if ($result2->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(FindInvalidEntries) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get all Reference Numbers of Transactions between $Start_Date and $End_Date from the Ledger.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }

  for ($i=0, $j=0; $i < $result2->ntuples; $i++)
  {
    $ref_num = $result2->getvalue($i, 0);

    # get the sum of all debits and all credits for this ref_num and verify that they equal
    $result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND parent_num = $ref_num AND dorc = true"); # all debits
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(FindInvalidEntries) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get sum of all Debits for Ref_Num = '$ref_num' from the Ledger.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $debit = $result->getvalue(0, 0);

    $result = $conn->exec("SELECT SUM(amount) FROM debitcredit_tb WHERE ledger_code = '$ledger_code' AND parent_num = $ref_num AND dorc = false"); # all credits
    if ($result->resultStatus != PGRES_TUPLES_OK)
    {
      $self->{ErrorString} = "<h1>(FindInvalidEntries) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get sum of all Credits for Ref_Num = '$ref_num' from the Ledger.<br>$sys_admin";
      $self->{ErrorOccurred} = 1;
      return ();
    }
    $credit = $result->getvalue(0, 0);

    if (not ($debit eq $credit))
    {
      $Results[$j++] = $ref_num;
    }
  }

  return @Results;
}

# GetCurrentWorkingYear
#    Takes  : Nothing
#    Returns: The current Fiscal Year from the Ledger.
sub GetCurrentWorkingYear
{
  my $self = shift;
  my $Result;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    $self->{ErrorString} = "(GetCurrentWorkingYear) Error:  This function does not take any values!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $result);

  # get the current fiscal year from the variables_tb table.
  $result = $conn->exec("SELECT working_year('$ledger_code')");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetCurrentWorkingYear) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get current fiscal year from ledger.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  if ($result->ntuples == 0)
  {
    $self->{ErrorString} = "<h1>(GetCurrentWorkingYear) Error:  Program not configured!</h1>\nThe Current Fiscal Year is not defined!<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }

  $Result = $result->getvalue(0, 0);

  return $Result;
}

# GetFiscalYearsBefore
#    Takes  : Fiscal Year to use
#    Returns: Array of Fiscal Years that are before the Fiscal Year specified.
sub GetFiscalYearsBefore
{
  my $self = shift;
  my @Results = ();
  my $FY;
  my $conn = $self->getGlobalDB;
  my $ledger_code = $self->getLedgerCode;

  if (defined $_[0])
  {
    $FY = shift;
  }
  else
  {
    $self->{ErrorString} = "(GetFiscalYearsBefore) Error:  You must specify a Fiscal Year to work with!\n";
    $self->{ErrorOccurred} = 1;
  }

  if ($self->{ErrorOccurred})
  {
    return ();
  }

  my ($i, $result);

  # verify that the fiscal year given is in the known_years_tb table.
  $result = $conn->exec("SELECT year FROM known_years_tb WHERE ledger_code = '$ledger_code' AND year = $FY");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetFiscalYearsBefore) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to verify that Fiscal Year = '$FY' is in the Ledger.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }
  if ($result->ntuples != 1)
  {
    $self->{ErrorString} = "<h1>(GetFiscalYearsBefore) Error:  Fiscal Year '$FY' not known!</h1>\nTried to verify that Fiscal Year = '$FY' is in the Ledger.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }

  # get the list of all fiscal years < the fiscal year specified.
  $result = $conn->exec("SELECT year FROM known_years_tb WHERE ledger_code = '$ledger_code' AND year < $FY");
  if ($result->resultStatus != PGRES_TUPLES_OK)
  {
    $self->{ErrorString} = "<h1>(GetFiscalYearsBefore) Error:  Accessing database failed!</h1>\n" . $conn->errorMessage . "<br>Tried to get all Fiscal Years < Fiscal Year = '$FY'.<br>$sys_admin";
    $self->{ErrorOccurred} = 1;
    return ();
  }

  for ($i=0; $i < $result->ntuples; $i++)
  {
    $Results[$i] = $result->getvalue($i, 0);
  }

  return @Results;
}

1;
